/****************************************************************************/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <asm/types.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <linux/sockios.h>
#include <sys/time.h>




#include "commun.h"

typedef struct t_context_fd
{
  int fd;
  int auto_state;
  int len_waited_for;
  int len_already_read;
} t_context_fd;

static t_context_fd g_ctx1;
static t_context_fd g_ctx2;
static int g_time_count;

/****************************************************************************/
static void fd_cloexec(int fd)
{
  int flags = fcntl(fd, F_GETFD);
  if (flags == -1)
    KOUT(" ");
  flags |= FD_CLOEXEC;
  if (fcntl(fd, F_SETFD, flags) == -1)
    KOUT(" ");
}
/*--------------------------------------------------------------------------*/

/*****************************************************************************/
static void set_sock_rx_buf(int fd)
{
  unsigned int datalen = 10000;
  if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF,
                 (char *)&datalen, sizeof(datalen)) < 0)
    KOUT(" ");
}
/*--------------------------------------------------------------------------*/


/****************************************************************************/
static time_t my_second(void)
{
  static time_t offset = 0; 
  struct timeval tv;
  gettimeofday(&tv, NULL);
  if (offset == 0)
    offset = tv.tv_sec;
  return (tv.tv_sec - offset);
}
/*--------------------------------------------------------------------------*/





/****************************************************************************/
static void nonblock(int fd)
{
  int flags = fcntl(fd, F_GETFL, 0);
  if (flags == -1)
    KOUT(" ");
  flags |= O_NONBLOCK|O_NDELAY;
  if (fcntl(fd, F_SETFL, flags) == -1)
    KOUT(" ");
}
/*--------------------------------------------------------------------------*/

/*****************************************************************************/
static void nonnonblock(int fd)
{
  int flags = fcntl(fd, F_GETFL, 0);
  flags &= ~O_NONBLOCK;
  flags &= ~O_NDELAY;
  if (fcntl(fd, F_SETFL, flags) == -1)
    KOUT(" ");
}
/*---------------------------------------------------------------------------*/




/****************************************************************************/
int util_socket_listen_inet( __u16 port )
{
  struct sockaddr s_addr;
  int fd;
  int optval=1;

  memset(&s_addr, 0, sizeof(s_addr));
  s_addr.sa_family = AF_INET;
  ((struct sockaddr_in*)&s_addr)->sin_addr.s_addr = htonl(INADDR_ANY);
  ((struct sockaddr_in*)&s_addr)->sin_port = htons( port );

  fd = socket( s_addr.sa_family, SOCK_STREAM, 0 );
  if ( fd <= 0 )
    KOUT("%d", errno);
  if (setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,&optval,sizeof(optval))<0)
    KOUT(" ");
  if ( bind(fd, &s_addr,sizeof(struct sockaddr_in) ) < 0 )
    {
    close(fd);
    fd = -1;
    }
  else
    {
    if ( listen(fd, 50) < 0 )
      {
      close(fd);
      fd = -1;
      }
    }
  if (fd > 0)
    {
    fd_cloexec(fd);
    }
  return(fd);
}
/*--------------------------------------------------------------------------*/

/****************************************************************************/
static void tx_to_dst(int fd_dst, int len, char *buf)
{
  static char tx[2000];
  memcpy (tx+4, buf, len);
  tx[0] = 0;
  tx[1] = 0;
  tx[2] = ((len & 0xFF00) >> 8) & 0xFF;
  tx[3] = len & 0xFF;
  write(fd_dst, tx, len+4);
}
/*--------------------------------------------------------------------------*/

/****************************************************************************/
static void process_fd_event(t_context_fd *src, t_context_fd *dst)

{
  static char rx[2000];
  int len;
  switch (src->auto_state)
    {
    case 0:
      len = read(src->fd, rx, 4);
      if (len != 4)
        {
        if ((len<0) || (len>4))
          KOUT("%d", len);
        nonnonblock(src->fd);
        len += read(src->fd, rx+len, 4-len);
        nonblock(src->fd);
        if (len != 4)
          KOUT("%d", len);
        }
      if ((rx[0] != (char) 0) || (rx[1] != (char) 0))
        KOUT(" ");
      src->len_waited_for =((rx[2] & 0xFF) << 8) + (rx[3] & 0xFF);
      if (src->len_waited_for == 0)
        KOUT(" ");
      src->len_already_read = 0;
      src->auto_state = 1;
      break;
    case 1:
      len = read(src->fd, rx, src->len_waited_for);
      if (len < 0)
        KOUT("%d", len);
      src->len_already_read += len;
      src->len_waited_for -= len;
      if (src->len_waited_for == 0)
        {
        src->auto_state = 0;
//        printf("RX: len:%d, %02X %02X %02X ...\n",
//               src->len_already_read, rx[0]&0xFF, rx[1]&0xFF, rx[2]&0xFF);
        tx_to_dst(dst->fd, src->len_already_read, rx);
        }
      break;
    default:
      KOUT(" ");
    }
};
/*--------------------------------------------------------------------------*/

/****************************************************************************/
static void select_wait_and_switch(t_context_fd *ctx1, t_context_fd *ctx2)
{
  fd_set infd;
  int result, fd_max;
  time_t cur_sec;
  static struct timeval timeout;
  FD_ZERO(&infd);
  FD_SET(ctx1->fd, &infd);
  FD_SET(ctx2->fd, &infd);
  fd_max = ctx1->fd;
  if (ctx2->fd > fd_max)
    fd_max = ctx2->fd;
  result = select(fd_max + 1, &infd, NULL, NULL, &timeout);
  if ( result < 0 )
    KOUT(" ");
  else if (result == 0) 
    {
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    }
  else 
    {
    if (FD_ISSET(ctx1->fd, &infd))
      process_fd_event(ctx1, ctx2);
    if (FD_ISSET(ctx2->fd, &infd))
      process_fd_event(ctx2, ctx1);
    }
  cur_sec = my_second();
  if (cur_sec != g_time_count)
    {
    g_time_count = cur_sec;
    printf("%d\n", g_time_count);
    if (!(g_time_count % 30))
      {
      printf("STOPING READING 5 sec\n");
      sleep(5);
      }
    }
}
/*--------------------------------------------------------------------------*/

/****************************************************************************/
void util_fd_accept(int fd_listen, int *fd)
{
  struct sockaddr s_addr;
  socklen_t sz = sizeof(s_addr);
  *fd = accept( fd_listen, &s_addr, &sz );
  if (*fd > 0)
    {
    nonblock(*fd);
    fd_cloexec(*fd);
    set_sock_rx_buf(*fd);
    }
  else
    KOUT(" ");
}
/*--------------------------------------------------------------------------*/



/****************************************************************************/
static int select_wait_listening(int fd)
{
  int result, new_fd;
  fd_set infd;
  FD_ZERO(&infd);
  FD_SET(fd, &infd);
  result = select(fd + 1, &infd, NULL, NULL, NULL);
  if ( result <= 0 )
    KOUT(" ");
  else
    {
    if (FD_ISSET(fd, &infd))
      util_fd_accept(fd, &new_fd);
    else
      KOUT(" ");
    }
  return new_fd;
}
/*--------------------------------------------------------------------------*/


/****************************************************************************/
int main(int argc, char *argv[])
{
  int fd_listen1, fd_listen2;
  g_time_count = 0;
  memset(&g_ctx1, 0, sizeof(t_context_fd));
  memset(&g_ctx2, 0, sizeof(t_context_fd));
  fd_listen1 = util_socket_listen_inet(47654);
  fd_listen2 = util_socket_listen_inet(47655);
  g_ctx1.fd = select_wait_listening(fd_listen1);
  printf("First connection\n");
  g_ctx2.fd = select_wait_listening(fd_listen2);
  printf("Second connection\n");
  for (;;)
    select_wait_and_switch(&g_ctx1, &g_ctx2);
  return 0;
}
/*--------------------------------------------------------------------------*/
